@@ -16,7 +16,7 @@
 
 import contextlib
 
-from typing import Any
+from typing import Any, Union
 
 _config = dict(
     strict_shape_check=True,
@@ -25,11 +25,16 @@ _config = dict(
     force_const_float64_to_bfloat16=False,
     convert_custom_gradient=True,
     infer_relu_from_jax2tf=True,
+    infer_cumulative_reduction_from_jax2tf=True,
     raise_on_prevent_gradient=True,
+    # TensorFlow uses high/highest precision on TPU.
+    resize_bilinear_precision="highest",
+    # This is temporary until checkify supports debug mode.
+    enable_checkify_for_asserts=False,
 )
 
 
-def get_config(name: str) -> bool:
+def get_config(name: str) -> Union[bool, str]:
   return _config[name]
 
 
