@@ -16,11 +16,13 @@
 
 import dataclasses
 import functools
-from typing import Any, Callable, Optional, Mapping, Sequence, Set, Tuple
+from typing import Any, Callable, List, Optional, Mapping, Sequence, Set, Tuple, Union
 
 from absl import logging
 
 import jax
+from jax._src.lax import control_flow as lax_control_flow
+from jax.experimental import checkify
 import jax.numpy as jnp
 import numpy as np
 import tensorflow as tf
@@ -77,6 +79,7 @@ _jax_ops = {
     "BitwiseXor": _get_jax_op(jnp.bitwise_xor, {"T"}),
     "BroadcastTo": _get_jax_op(anp.broadcast_to, {"T", "Tidx"}),
     "Ceil": _get_jax_op(jnp.ceil, {"T"}),
+    "Cholesky": _get_jax_op(jnp.linalg.cholesky, {"T"}),
     "Complex": _get_jax_op(jax.lax.complex, {"T", "Tout"}),
     "ComplexAbs": _get_jax_op(jax.lax.abs, {"T", "Tout"}),
     "Conj": _get_jax_op(jax.lax.conj, {"T", "Tout"}),
@@ -91,7 +94,6 @@ _jax_ops = {
     "Erfinv": _get_jax_op(jax.lax.erf_inv, {"T"}),
     "Exp": _get_jax_op(jnp.exp, {"T"}),
     "Expm1": _get_jax_op(jnp.expm1, {"T"}),
-    "ExpandDims": _get_jax_op(anp.expand_dims, {"T", "Tdim"}),
     "Floor": _get_jax_op(jnp.floor, {"T"}),
     "FloorMod": _get_jax_op(anp.mod, {"T"}),
     "FloorDiv": _get_jax_op(anp.floor_divide, {"T"}),
@@ -103,6 +105,7 @@ _jax_ops = {
     "Imag": _get_jax_op(jax.lax.imag, {"T", "Tout"}),
     "IsFinite": _get_jax_op(jnp.isfinite, {"T"}),
     "Invert": _get_jax_op(jnp.bitwise_not, {"T"}),
+    "InvertPermutation": _get_jax_op(anp.invert_permutation, {"T"}),
     "L2Loss": _get_jax_op(lambda x: 0.5 * jnp.sum(jnp.square(x)), {"T"}),
     "LeftShift": _get_jax_op(jnp.left_shift, {"T"}),
     "Less": _get_jax_op(anp.less, {"T", "incompatible_shape_error"}),
@@ -146,7 +149,21 @@ _jax_ops = {
     "Sub": _get_jax_op(anp.subtract, {"T"}),
     "Tan": _get_jax_op(jnp.tan, {"T"}),
     "Tanh": _get_jax_op(jnp.tanh, {"T"}),
+    "TensorListFromTensor": _get_jax_op(
+        lambda xs, element_shape: xs, {"element_dtype", "shape_type"}),
     "Tile": _get_jax_op(anp.tile, {"T", "Tmultiples"}),
+    "UnsortedSegmentMax": _get_jax_op(
+        functools.partial(jax.ops.segment_max, indices_are_sorted=False),
+        {"T", "Tindices", "Tnumsegments"}),
+    "UnsortedSegmentMin": _get_jax_op(
+        functools.partial(jax.ops.segment_min, indices_are_sorted=False),
+        {"T", "Tindices", "Tnumsegments"}),
+    "UnsortedSegmentProd": _get_jax_op(
+        functools.partial(jax.ops.segment_prod, indices_are_sorted=False),
+        {"T", "Tindices", "Tnumsegments"}),
+    "UnsortedSegmentSum": _get_jax_op(
+        functools.partial(jax.ops.segment_sum, indices_are_sorted=False),
+        {"T", "Tindices", "Tnumsegments"}),
     "Where": _get_jax_op(jnp.argwhere, {"T"}),
     "ZerosLike": _get_jax_op(jnp.zeros_like, {"T"}),
     # The assignment logic is handled in _OpNode and convert().
@@ -154,6 +171,12 @@ _jax_ops = {
     "AssignSubVariableOp": _get_jax_op(jnp.subtract, {"dtype"}),
     "AssignVariableOp": _get_jax_op(
         lambda var, x: x, {"dtype", "validate_shape"}),
+    "RFFT":
+        _get_jax_op(lambda x, a: jnp.fft.rfft(x, a[0]), {"Tcomplex", "Treal"}),
+    "IRFFT":
+        _get_jax_op(lambda x, a: jnp.fft.irfft(x, a[0]), {"Tcomplex", "Treal"}),
+    "Angle":
+        _get_jax_op(jnp.angle, {"T", "Tout"}),
 }
 
 
@@ -212,7 +235,7 @@ def _assert(proto):
 
   logging.warning("Assert has no effect.")
 
-  return lambda cond, data: _EMPTY_RETURN_VALUE
+  return lambda cond, *data: _EMPTY_RETURN_VALUE
 
 
 @register_operation("AvgPool")
@@ -261,7 +284,7 @@ def _bias_add(proto):
   if data_format == "NHWC":
     expand_axis_fn = lambda x: [d for d in range(x.ndim) if d != x.ndim - 1]
   elif data_format == "NCHW":
-    # TODO(shaobohou) this seems wrong but matches TF behaviour.
+    # TODO(b/266553458) this seems wrong but matches TF behaviour.
     expand_axis_fn = lambda x: [d for d in range(x.ndim) if d != 1]
   else:
     raise ValueError(f"Found unsupported data format {data_format}.")
@@ -333,6 +356,26 @@ def _cast(proto):
   return _func
 
 
+@register_operation("CheckNumerics")
+def _check_numerics(proto):
+  """Parse an CheckNumerics Op."""
+  _check_attrs(proto, {"T", "message"})
+
+  message = str(proto.attr["message"].s, "utf-8")
+
+  def _func(operand: jnp.ndarray) -> jnp.ndarray:
+    if config.get_config("enable_checkify_for_asserts"):
+      checkify.check(
+          jnp.logical_not(jnp.any(jnp.isnan(operand))),
+          f"{message} : Found NaN values.")
+      checkify.check(
+          jnp.logical_not(jnp.any(jnp.isinf(operand))),
+          f"{message} : Found Inf values.")
+    return operand
+
+  return _func
+
+
 @register_operation("ConjugateTranspose")
 def _conjugate_transpose(proto):
   _check_attrs(proto, {"T", "Tperm"})
@@ -546,6 +589,110 @@ def _depthwise_conv2d(proto):
   return _func
 
 
+def _div_no_nan_forward(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
+  """Computes a safe divide which returns 0 if y (denominator) is zero."""
+  div = anp.divide(x, y)
+  return jnp.where(y == 0.0, jnp.zeros_like(div), div)
+
+
+@register_operation("DivNoNan")
+def _div_no_nan(proto):
+  """Parse a DivNoNan op."""
+  _check_attrs(proto, {"T"})
+
+  @jax.custom_gradient
+  def _func(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
+
+    def _grad(g: jnp.ndarray) -> jnp.ndarray:
+      """Given upstream grad G and a Div op: Z = X/Y, the gradients are.
+
+        dX = G / Y
+        dY = -G*X / Y^2 = (X/Y) * -G / Y = -G*Z / Y
+
+      Following tensorflow/tensorflow/c/experimental/gradients/math_grad.cc
+      for handling NaNs.
+
+      Args:
+        g: Upstream gradient.
+
+      Returns:
+        forward value: div_no_nan(x, y)
+        grad: Gradient information in TF format.
+      """
+      output = _div_no_nan_forward(x, y)
+      dx = _div_no_nan_forward(g, y)
+      dy = _div_no_nan_forward(-g * output, y)
+      return dx, dy
+
+    return _div_no_nan_forward(x, y), _grad
+
+  return _func
+
+
+@register_operation("Eig")
+def _eig(proto):
+  """Parse an Eig Op."""
+  _check_attrs(proto, {"T", "Tout", "compute_v"})
+
+  compute_v = proto.attr["compute_v"].b
+
+  def _func(x: jnp.ndarray) -> jnp.ndarray:
+    if compute_v:
+      evals, evecs = jnp.linalg.eig(x)
+    else:
+      evals = jnp.linalg.eigvals(x)
+      evecs = jnp.zeros(shape=(), dtype=evals.dtype)
+
+    # Sorting by eigenvalues to match tf.raw_ops.Eig better.
+    sort_fn = lambda vals, inds: vals[..., inds]
+    for _ in range(len(x.shape) - 2):
+      sort_fn = jax.vmap(sort_fn, in_axes=(0, 0))
+    einds = jnp.argsort(jnp.absolute(evals), axis=-1)
+    evals = sort_fn(evals, einds)
+
+    if compute_v:
+      evecs = sort_fn(evecs, einds)
+
+    return evals, evecs
+
+  return _func
+
+
+@register_operation("SelfAdjointEigV2")
+def _eigh(proto):
+  """Parse an SelfAdjointEigV2 Op."""
+  _check_attrs(proto, {"T", "compute_v"})
+
+  compute_v = proto.attr["compute_v"].b
+
+  def symmetrize(x):
+    return jnp.tril(x) + jnp.swapaxes(jnp.tril(x, -1), -2, -1)
+
+  def _func(x: jnp.ndarray) -> jnp.ndarray:
+    if compute_v:
+      evals, evecs = jnp.linalg.eigh(x, symmetrize_input=False)
+    else:
+      # symmetrize_input does not exist for eigvalsh.
+      # See https://github.com/google/jax/issues/9473
+      evals, evecs = jnp.linalg.eigvalsh(symmetrize(x)), None
+
+    # Sorting by eigenvalues to tf.raw_ops.Eig better.
+    sort_fn = lambda vals, inds: vals[..., inds]
+    for _ in range(len(x.shape) - 2):
+      sort_fn = jax.vmap(sort_fn, in_axes=(0, 0))
+    einds = jnp.argsort(evals, axis=-1)
+    evals = sort_fn(evals, einds)
+    if compute_v:
+      evecs = sort_fn(evecs, einds)
+
+    if compute_v:
+      return evals, evecs
+    else:
+      return evals, jnp.zeros(shape=(), dtype=evals.dtype)
+
+  return _func
+
+
 @register_operation("Einsum")
 def _einsum(proto):
   """Parse an Einsum Op."""
@@ -577,6 +724,36 @@ def _empty(proto):
   return _func
 
 
+@register_operation("EnsureShape")
+def _ensure_shape(proto):
+  """Parse an EnsureShape Op."""
+  _check_attrs(proto, {"T", "shape"})
+
+  shape = [dim.size for dim in proto.attr["shape"].shape.dim]
+
+  def is_compatible(x, s):
+    return x == -1 or s == -1 or x == s
+
+  def _func(x: jnp.ndarray) -> jnp.ndarray:
+    if (len(x.shape) != len(shape) or
+        not all(is_compatible(x, s) for x, s in zip(x.shape, shape))):
+      raise ValueError(f"Expected shape={shape}, found {x.shape}.")
+    return x
+
+  return _func
+
+
+@register_operation("ExpandDims")
+def _expand_dims(proto):
+  """Parse an ExpandDims Op."""
+  _check_attrs(proto, {"T", "Tdim"})
+
+  def _func(arr: jnp.ndarray, axis: jnp.ndarray) -> jnp.ndarray:
+    return anp.expand_dims(arr, axis=axis.tolist())
+
+  return _func
+
+
 @register_operation("Fill")
 def _fill(proto):
   """Parse an Fill op."""
@@ -666,19 +843,24 @@ def _gather_nd(proto):
   return _func
 
 
+@register_operation("ResourceGather")
 @register_operation("GatherV2")
 def _gather(proto):
   """Parse a GatherV2 Op."""
-  _check_attrs(proto, {"Taxis", "Tindices", "Tparams", "batch_dims"})
+  _check_attrs(proto, {
+      "Taxis", "Tindices", "Tparams", "batch_dims", "dtype", "validate_indices"
+  })
 
   batch_dims = proto.attr["batch_dims"].i
   if batch_dims < 0:
     raise ValueError(f"batch_dims={batch_dims} must be non-negative.")
 
+  # TODO(b/249826984) Add test for ResourceGather, dtype and validate_indices.
+
   def _func(
       params: jnp.ndarray,
       indices: jnp.ndarray,
-      axis: jnp.ndarray,
+      axis: jnp.ndarray = np.array(0, dtype=np.int32),
   ) -> jnp.ndarray:
     return anp.gather(
         params, indices, axis=axis.tolist(), batch_dims=batch_dims)
@@ -799,7 +981,7 @@ def _batch_matmul(proto):
   adj_x = proto.attr["adj_x"].b
   adj_y = proto.attr["adj_y"].b
 
-  # TODO(shaobohou) Add test for arrays with complex values.
+  # TODO(b/266553251) Add test for arrays with complex values.
   def _func(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
     if adj_x:
       x = jnp.conjugate(jnp.swapaxes(x, -1, -2))
@@ -865,6 +1047,26 @@ def _matrix_band_part(proto):
   return _func
 
 
+@register_operation("MatrixTriangularSolve")
+def _matrix_triangular_solve(proto):
+  """Parse an MatrixTriangularSolve Op."""
+  _check_attrs(proto, {"T", "lower", "adjoint"})
+
+  lower = proto.attr["lower"].b
+  adjoint = proto.attr["adjoint"].b
+
+  def _func(arr: jnp.ndarray, rhs: jnp.ndarray) -> jnp.ndarray:
+    return jax.lax.linalg.triangular_solve(
+        arr,
+        rhs,
+        lower=lower,
+        transpose_a=adjoint,
+        conjugate_a=adjoint,
+        left_side=True)
+
+  return _func
+
+
 @register_operation("Max")
 def _max(proto):
   _check_attrs(proto, {"T", "Tidx", "keep_dims"})
@@ -999,7 +1201,6 @@ class _PartitionedCall(_HigherOrderFunction):
     return inner_fn(*args, rng=rng)
 
 
-# TODO(shaobohou) Add test for StatefulPartitionedCall.
 @register_operation("StatefulPartitionedCall")
 @register_operation("PartitionedCall")
 def _partitioned_call(proto):
@@ -1009,7 +1210,7 @@ def _partitioned_call(proto):
 
   inner_fn = proto.attr["f"].func.name
   op_config = str(proto.attr["config"].s, "utf-8")
-  op_config_proto = proto.attr["config_proto"].s  # TODO(shaobohou) decode this?
+  op_config_proto = proto.attr["config_proto"].s
   executor_type = str(proto.attr["executor_type"].s, "utf-8")
   del op_config, op_config_proto, executor_type
 
@@ -1066,6 +1267,19 @@ def _prod(proto):
   return _func
 
 
+@register_operation("Qr")
+def _qr(proto):
+  """Parse an QR Op."""
+  _check_attrs(proto, {"T", "full_matrices"})
+
+  full_matrices = proto.attr["full_matrices"].b
+
+  def _func(x: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
+    return jnp.linalg.qr(x, mode="complete" if full_matrices else "reduced")
+
+  return _func
+
+
 @register_operation("RandomStandardNormal")
 def _random_standard_normal(proto):
   """Parse a RandomStandardNormal op."""
@@ -1148,7 +1362,7 @@ def _reshape(proto):
 
 
 @register_operation("ResizeBilinear")
-def _resize_linear(proto):
+def _resize_bilinear(proto):
   """Parse a ResizeBilinear op."""
   _check_attrs(proto, {"T", "align_corners", "half_pixel_centers"})
 
@@ -1189,12 +1403,58 @@ def _resize_linear(proto):
         translation=translation,
         method="linear",
         antialias=False,
-        precision=None,
+        precision=config.get_config("resize_bilinear_precision"),
     )
 
   return _func
 
 
+@register_operation("ResizeNearestNeighbor")
+def _resize_nearest_neighbor(proto):
+  """Parse a ResizeNearestNeighbor op."""
+  _check_attrs(proto, {"T", "align_corners", "half_pixel_centers"})
+
+  align_corners = proto.attr["align_corners"].b
+  half_pixel_centers = proto.attr["half_pixel_centers"].b
+
+  if align_corners or not half_pixel_centers:
+    # Not supported by tf.raw_ops.ResizeNearestNeighbor.
+    raise ValueError(
+        "Only align_corners=False and half_pixel_centers=True is supported.")
+
+  def _func(images: jnp.ndarray, size: jnp.ndarray) -> jnp.ndarray:
+    if len(images.shape) != 4:
+      raise ValueError(
+          "Expected A 4D tensor with shape [batch, height, width, channels], "
+          f"found {images.shape}")
+
+    inp_batch, _, _, inp_channels = images.shape
+    out_height, out_width = size.tolist()
+
+    return jax.image.resize(
+        images,
+        shape=(inp_batch, out_height, out_width, inp_channels),
+        method=jax.image.ResizeMethod.NEAREST,
+    )
+
+  return _func
+
+
+@register_operation("Roll")
+def _roll(proto):
+  """Parse a Roll Op."""
+  _check_attrs(proto, {"T", "Tshift", "Taxis"})
+
+  def _func(
+      operand: jnp.ndarray,
+      shift: jnp.ndarray,
+      axis: jnp.ndarray,
+  ) -> jnp.ndarray:
+    return anp.roll(operand, shift=shift.tolist(), axis=axis.tolist())
+
+  return _func
+
+
 @register_operation("ScatterNd")
 def _scatter_nd(proto):
   """Parse a ScatterNd op."""
@@ -1318,6 +1578,14 @@ def _squeeze(proto):
   return lambda x: anp.squeeze(x, axis=axis)
 
 
+@register_operation("StatelessRandomGetAlg")
+def _stateless_random_get_alg(proto):
+  _check_attrs(proto, set({}))
+  # Only jax.lax.rng_bit_generator allows users to choose the algorithm.
+  # So this will most likely be ignored.
+  return lambda: tf.random.Algorithm.AUTO_SELECT.value
+
+
 @register_operation("StatelessRandomGetKeyCounter")
 def _stateless_random_get_key_counter(proto):
   _check_attrs(proto, {"T", "Tseed"})
@@ -1366,7 +1634,7 @@ def _stateless_random_normal_v2(proto):
       counter: jnp.ndarray,
       alg: jnp.ndarray,
   ) -> jnp.ndarray:
-    del counter, alg  # TODO(shaobohou) combine key and counter?
+    del counter, alg  # TODO(b/266553394) combine key and counter?
     return jax.random.normal(key=key, shape=shape, dtype=jax_dtype)
 
   return _func
@@ -1386,7 +1654,7 @@ def _stateless_random_uniform_v2(proto):
       counter: jnp.ndarray,
       alg: jnp.ndarray,
   ) -> jnp.ndarray:
-    del counter, alg  # TODO(shaobohou) combine key and counter?
+    del counter, alg  # TODO(b/266553394) combine key and counter?
     return jax.random.uniform(key=key, shape=shape, dtype=jax_dtype)
 
   return _func
@@ -1409,7 +1677,7 @@ def _stateless_random_uniform_int_v2(proto):
       minval: jnp.ndarray = jnp.iinfo(jax_dtype).min,
       maxval: jnp.ndarray = jnp.iinfo(jax_dtype).max,
   ) -> jnp.ndarray:
-    del counter, alg  # TODO(shaobohou) combine key and counter?
+    del counter, alg  # TODO(b/266553394) combine key and counter?
     return jax.random.randint(
         key=key, shape=shape, minval=minval, maxval=maxval, dtype=jax_dtype,)
 
@@ -1472,6 +1740,18 @@ class _StatelessWhile(_HigherOrderFunction):
       outputs = tuple([outputs[idx] for idx in trace_idxs])
       return outputs + (key,)
 
+    def maybe_initialise(arg, spec):
+      # TensorList inputs can be uninitialized.
+      if isinstance(spec, jax.ShapeDtypeStruct) and arg.size == 0:
+        return jnp.zeros(shape=spec.shape, dtype=spec.dtype)
+      else:
+        return arg
+
+    output_specs = jax.eval_shape(real_body, trace_args + (rng,))
+    trace_args = tuple(
+        maybe_initialise(arg, spec)
+        for arg, spec in zip(trace_args, output_specs))
+
     outputs = jax.lax.while_loop(real_cond, real_body, trace_args + (rng,))
     *outputs, _ = outputs  # Drop rng.
     outputs = _merge_args(const_idxs_args, tuple(zip(trace_idxs, outputs)))
@@ -1485,8 +1765,6 @@ def _stateless_while(proto):
   """Parse a StatelessWhile op."""
   _check_attrs(proto,
                {"T", "body", "cond", "parallel_iterations", "output_shapes"})
-  # TODO(shaobohou) Check proto.arg_attr?
-
   body_name = proto.attr["body"].func.name
   cond_name = proto.attr["cond"].func.name
   parallel_iterations = proto.attr["parallel_iterations"].i
@@ -1547,13 +1825,14 @@ def _strided_slice(proto):
           x_dim = x.shape[dim]
           if (stride == 1 and
               jax.core.symbolic_equal_dim(beg_dim or 0, 0) and
-              jax.core.symbolic_equal_dim(end_dim or x_dim, x_dim)):
+              jax.core.symbolic_equal_dim(x_dim if end_dim is None else end_dim,
+                                          x_dim)):
             slices.append(slice(None, None, None))
           else:
             slices.append(slice(beg_dim, end_dim, stride))
         dim += 1
 
-    # TODO(shaobohou) Handle stride=1 slicing along polymoprhic dimensions.
+    # TODO(b/266553742) Handle stride=1 slicing along polymoprhic dimensions.
     return x[tuple(slices)]
 
   return _func
@@ -1571,6 +1850,143 @@ def _sum(proto):
   return _func
 
 
+@register_operation("Svd")
+def _svd(proto):
+  """Parse an Svd Op."""
+  _check_attrs(proto, {"T", "compute_uv", "full_matrices"})
+
+  compute_uv = proto.attr["compute_uv"].b
+  full_matrices = proto.attr["full_matrices"].b
+
+  def _func(x: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
+    res = jnp.linalg.svd(
+        x, compute_uv=compute_uv, full_matrices=full_matrices)
+    if compute_uv:
+      u, s, vh = res
+      v = jnp.conjugate(jnp.swapaxes(vh, -1, -2))
+    else:
+      u, s, v = None, res, None
+
+    return s, u, v
+
+  return _func
+
+
+@register_operation("TensorListGetItem")
+def _tensor_list_get_item(proto):
+  """Parse an TensorListGetItem Op."""
+  _check_attrs(proto, {"element_dtype"})
+
+  dtype = tf.as_dtype(proto.attr["element_dtype"].type)
+  dtype = anp.get_jax_dtype(dtype)
+
+  def _func(
+      xs: jnp.ndarray,
+      index: jnp.ndarray,
+      element_shape: jnp.ndarray,
+  ) -> jnp.ndarray:
+    assert xs.dtype == dtype
+    if xs.size == 0:
+      return np.zeros(element_shape, dtype=dtype)
+    if isinstance(index, jax.Array):
+      xs = jnp.array(xs)
+    return xs[index]
+
+  return _func
+
+
+def _create_tensor_list(
+    num_elements: int,
+    shape: Union[int, List[int]],
+    dtype: Any,
+) -> np.ndarray:
+  """Create a tensor corresponding to a stacked tensor list."""
+  if not isinstance(shape, list):
+    shape = [shape]
+  # shape can be unspecified as -1.
+  shape = [max(0, sz) for sz in shape]
+  return np.zeros([num_elements] + shape, dtype=dtype)
+
+
+@register_operation("TensorListReserve")
+def _tensor_list_reserve(proto):
+  """Parse an TensorListReserve Op."""
+  _check_attrs(proto, {"element_dtype", "shape_type"})
+
+  dtype = tf.as_dtype(proto.attr["element_dtype"].type)
+  dtype = anp.get_jax_dtype(dtype)
+
+  def _func(shape: jnp.ndarray, num_elements: jnp.ndarray) -> jnp.ndarray:
+    return _create_tensor_list(
+        num_elements.tolist(), shape.tolist(), dtype=dtype)
+
+  return _func
+
+
+@register_operation("TensorListSetItem")
+def _tensor_list_set_item(proto):
+  """Parse an TensorListSetItem Op."""
+  _check_attrs(proto, {"element_dtype", "resize_if_index_out_of_bounds"})
+
+  if proto.attr["resize_if_index_out_of_bounds"].b:
+    raise ValueError(
+        "resize_if_index_out_of_bounds=True in TensorListSetItem is not yet"
+        " supported."
+    )
+
+  dtype = tf.as_dtype(proto.attr["element_dtype"].type)
+  dtype = anp.get_jax_dtype(dtype)
+
+  def _func(
+      xs: jnp.ndarray,
+      index: jnp.ndarray,
+      item: jnp.ndarray,
+  ) -> jnp.ndarray:
+    assert xs.shape
+    if xs.size == 0:
+      xs = _create_tensor_list(xs.shape[0], list(item.shape), dtype=dtype)
+    if isinstance(index, jax.Array):
+      xs = jnp.array(xs)
+    return xs.at[index].set(item)
+
+  return _func
+
+
+@register_operation("TensorListStack")
+def _tensor_list_stack(proto):
+  """Parse an TensorListStack Op."""
+  _check_attrs(proto, {"element_dtype", "num_elements"})
+
+  dtype = tf.as_dtype(proto.attr["element_dtype"].type)
+  dtype = anp.get_jax_dtype(dtype)
+
+  def _func(xs: jnp.ndarray, shape: jnp.ndarray) -> jnp.ndarray:
+    if xs.size == 0:
+      xs = _create_tensor_list(xs.shape[0], shape.tolist(), dtype=dtype)
+    return xs
+
+  return _func
+
+
+@register_operation("TensorScatterUpdate")
+def _tensor_scatter_update(proto):
+  """Parse an TensorScatterUpdate Op."""
+  _check_attrs(proto, {"T", "Tindices"})
+
+  def _func(
+      operand: jnp.ndarray,
+      indices: jnp.ndarray,
+      updates: jnp.ndarray,
+  ) -> jnp.ndarray:
+    dimension_numbers = jax.lax.ScatterDimensionNumbers(
+        range(1, updates.ndim), range(indices.shape[-1]),
+        range(indices.shape[-1]))
+    return jax.lax.scatter(
+        operand, indices, updates, dimension_numbers=dimension_numbers)
+
+  return _func
+
+
 @register_operation("TopKV2")
 def _top_k(proto):
   _check_attrs(proto, {"T", "sorted"})
@@ -1730,6 +2146,19 @@ def _xla_gather(proto):
   return _func
 
 
+@register_operation("XlaOptimizationBarrier")
+def _xla_optimization_barrier(proto):
+  """Parse a XlaOptimizationBarrier op."""
+  _check_attrs(proto, {"T"})
+
+  def _func(*operands: jnp.ndarray) -> Tuple[jnp.ndarray, ...]:
+    # TODO(b/241584320) Note this does not reproduce the remat transform in the
+    # forward pass, which may require some heurstics when parsing the graphdef.
+    return lax_control_flow.optimization_barrier_p.bind(*operands)
+
+  return _func
+
+
 @register_operation("XlaPad")
 def _xla_pad(proto):
   """Parse a XlaPad op."""
@@ -1749,6 +2178,20 @@ def _xla_pad(proto):
   return _func
 
 
+@register_operation("XlaReducePrecision")
+def _xla_reduce_precision(proto):
+  """Parse a XlaReducePrecision op."""
+  _check_attrs(proto, {"T", "exponent_bits", "mantissa_bits"})
+
+  exponent = proto.attr["exponent_bits"].i
+  mantissa = proto.attr["mantissa_bits"].i
+
+  def _func(operand: jnp.ndarray) -> jnp.ndarray:
+    return jax.lax.reduce_precision(operand, exponent, mantissa)
+
+  return _func
+
+
 def _maybe_get_jaxpreqn(
     jaxpr: jax.core.ClosedJaxpr) -> Optional[jax.core.JaxprEqn]:
   def is_all_vars(vs):
@@ -1851,6 +2294,26 @@ def _xla_variadic_sort(proto):
   return _XlaVariadicSort(dict(comparator=comparator), is_stable=is_stable)
 
 
+# Taken from https://github.com/google/jax/blob/main/jax/_src/lax/lax.py#L1056
+def _get_max_identity(dtype):
+  if jax.dtypes.issubdtype(dtype, np.inexact):
+    return np.array(-np.inf, dtype)
+  elif jax.dtypes.issubdtype(dtype, np.integer):
+    return np.array(jax.dtypes.iinfo(dtype).min, dtype)
+  elif jax.dtypes.issubdtype(dtype, np.bool_):
+    return np.array(False, np.bool_)
+
+
+# Taken from https://github.com/google/jax/blob/main/jax/_src/lax/lax.py#L1064
+def _get_min_identity(dtype):
+  if jax.dtypes.issubdtype(dtype, np.inexact):
+    return np.array(np.inf, dtype)
+  elif jax.dtypes.issubdtype(dtype, np.integer):
+    return np.array(jax.dtypes.iinfo(dtype).max, dtype)
+  elif jax.dtypes.issubdtype(dtype, np.bool_):
+    return np.array(True, np.bool_)
+
+
 class _XlaReduceWindow(_HigherOrderFunction):
   """Represents a XlaReduceWindow Op."""
 
@@ -1866,11 +2329,18 @@ class _XlaReduceWindow(_HigherOrderFunction):
       *,
       computation: Callable[..., Any],
   ):
+    window_dimensions = window_dimensions.tolist()
+    window_strides = window_strides.tolist()
+    padding = padding.tolist()
+    base_dilation = base_dilation.tolist()
+    window_dilation = window_dilation.tolist()
+
     # Pattern matching computations that can be specialized.
     primitives = {
         jax.lax.max_p: jax.lax.max,
         jax.lax.min_p: jax.lax.min,
         jax.lax.add_p: jax.lax.add,
+        jax.lax.mul_p: jax.lax.mul,
     }
     computation_jaxpr = jax.make_jaxpr(computation)(init_value, init_value)
     computation_eqn = _maybe_get_jaxpreqn(computation_jaxpr)
@@ -1881,15 +2351,56 @@ class _XlaReduceWindow(_HigherOrderFunction):
       logging.info("Calling reduce_window with the following computation:\n%s",
                    computation_jaxpr)
 
-    return jax.lax.reduce_window(
-        operand,
-        init_value,
-        computation=computation_fn,
-        window_dimensions=window_dimensions.tolist(),
-        window_strides=window_strides.tolist(),
-        padding=[tuple(v) for v in padding.tolist()],
-        base_dilation=base_dilation.tolist(),
-        window_dilation=window_dilation.tolist())
+    def infer_cumulative_reduction():
+      ndims = len(window_dimensions)
+      assert ndims == operand.ndim
+      reduce_axis = np.argmax(window_dimensions)
+      reduce_dim = operand.shape[reduce_axis]
+      dims = [1] * ndims
+      dims[reduce_axis] = reduce_dim
+      if not (window_dimensions == dims and window_strides == [1] * ndims and
+              base_dilation == [1] * ndims and window_dilation == [1] * ndims):
+        return (None, None, None)
+
+      # Determine direction of reduction.
+      normal_padding = [[0, 0]] * ndims
+      normal_padding[reduce_axis] = [reduce_dim - 1, 0]
+      reverse_padding = [[0, 0]] * ndims
+      reverse_padding[reduce_axis] = [0, reduce_dim - 1]
+      reverse = None
+      if padding == normal_padding:
+        reverse = False
+      elif padding == reverse_padding:
+        reverse = True
+      else:
+        return (None, None, None)
+
+      if computation_fn is jax.lax.add and init_value == 0:
+        return (jax.lax.cumsum, reduce_axis, reverse)
+      elif computation_fn is jax.lax.mul and init_value == 1:
+        return (jax.lax.cumprod, reduce_axis, reverse)
+      elif (computation_fn is jax.lax.max and
+            init_value == _get_max_identity(operand.dtype)):
+        return (jax.lax.cummax, reduce_axis, reverse)
+      elif (computation_fn is jax.lax.min and
+            init_value == _get_min_identity(operand.dtype)):
+        return (jax.lax.cummin, reduce_axis, reverse)
+      else:
+        return (None, None, None)
+
+    reducer, reduce_axis, reverse = infer_cumulative_reduction()
+    if reducer and config.get_config("infer_cumulative_reduction_from_jax2tf"):
+      return reducer(operand, axis=reduce_axis, reverse=reverse)
+    else:
+      return jax.lax.reduce_window(
+          operand,
+          init_value,
+          computation=computation_fn,
+          window_dimensions=window_dimensions,
+          window_strides=window_strides,
+          padding=[tuple(v) for v in padding],
+          base_dilation=base_dilation,
+          window_dilation=window_dilation)
 
 
 @register_operation("XlaReduceWindow")
@@ -2012,7 +2523,7 @@ class _XlaSelectAndScatter(_HigherOrderFunction):
       raise ValueError(
           f"Only Add is supported as scatter function, found {scatter_jaxpr}.")
 
-    # TODO(shaobohou) Support jax.lax.add for AvgPool.
+    # TODO(b/266553393) Support jax.lax.add for AvgPool.
     select_primitives = {
         jax.lax.ge_p: (-jnp.inf, jax.lax.max),
         jax.lax.le_p: (jnp.inf, jax.lax.min),
