@@ -20,10 +20,13 @@ from absl.testing import parameterized
 
 import chex
 import jax
+from jax.experimental import checkify
 import numpy as np
 
 import tensorflow as tf
+from tf2jax._src import config
 from tf2jax._src import ops
+from tf2jax._src import test_util
 from tf2jax._src import tf2jax
 import tree
 
@@ -32,12 +35,7 @@ def _reorder(vals, inds):
   return [vals[idx] for idx in inds]
 
 
-@contextlib.contextmanager
-def _nullcontext(enter_result=None):
-  yield enter_result
-
-
-class OpsTest(tf.test.TestCase, parameterized.TestCase):
+class OpsTest(test_util.TestCase):
 
   def test_get_unsupported(self):
     unsupported = ops.get_unsupported_operations(
@@ -46,18 +44,23 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
 
   def _assert_if_jitted(self, err):
     jitted = self.variant.type == chex.ChexVariantType.WITH_JIT
-    return self.assertRaises(err) if jitted else _nullcontext()
+    return self.assertRaises(err) if jitted else contextlib.nullcontext()
 
   def _test_convert(self,
                     tf_func,
                     inputs,
+                    *,
                     check_shape_only=False,
-                    functional=True):
+                    functional=True,
+                    atol=1e-5):
     if not isinstance(inputs, (list, tuple)):
       inputs = (inputs,)
 
+    if not hasattr(tf_func, "get_concrete_function"):
+      tf_func = tf.function(tf_func, jit_compile=True)
+
     jax_func, jax_params = tf2jax.convert(
-        tf.function(tf_func), *tree.map_structure(np.zeros_like, inputs))
+        tf_func, *tree.map_structure(np.zeros_like, inputs))
     if functional:
       self.assertEmpty(jax_params, "Expected no parameters for pure Ops.")
 
@@ -72,7 +75,7 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
         tree.flatten(tf_results), tree.flatten(jax_results)):
       self.assertEqual(tf_res.shape, jax_res.shape)
       if not check_shape_only:
-        self.assertAllClose(np.asarray(tf_res), jax_res, atol=1e-5)
+        self.assertAllClose(np.asarray(tf_res), jax_res, atol=atol)
 
     return jax_results, new_jax_params
 
@@ -82,7 +85,13 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
   def test_activations(self, op_name):
     np.random.seed(42)
     inputs = [np.random.normal(size=(10, 5)).astype(np.float32)]
-    self._test_convert(getattr(tf.nn, op_name), inputs)
+
+    if jax.default_backend().lower() == "tpu" and op_name in ("softplus",):
+      tols = dict(atol=1e-4)
+    else:
+      tols = {}
+
+    self._test_convert(getattr(tf.nn, op_name), inputs, **tols)
 
   @chex.variants(with_jit=True, without_jit=True)
   def test_assert(self):
@@ -135,7 +144,13 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
 
     def tf_func(x, y):
       return getattr(tf.raw_ops, op_name)(x=x, y=y)
-    self._test_convert(tf_func, inputs)
+
+    if jax.default_backend().lower() == "tpu" and op_name in ("Pow",):
+      tols = dict(atol=1e-4)
+    else:
+      tols = {}
+
+    self._test_convert(tf_func, inputs, **tols)
 
     # Check static inputs result in static outputs.
     def tf_static():
@@ -346,6 +361,43 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return tf.zeros(broadcast_to(inputs)[0])
     self._test_convert(broadcast_to_static, [])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      ("valid", np.array((3.14, 42.0), dtype=np.float32), None),
+      ("nan", np.array((np.nan, 42.0), dtype=np.float32), "Found NaN values"),
+      ("inf", np.array((3.14, np.inf), dtype=np.float32), "Found Inf values"),
+  )
+  def test_check_numerics(self, inputs, expected_message):
+
+    def check_numerics(x):
+      return tf.raw_ops.CheckNumerics(tensor=x, message="Checking")
+
+    with config.override_config("enable_checkify_for_asserts", False):
+      jax_fn = tf2jax.convert_functional(tf.function(check_numerics), inputs)
+      jax_fn = self.variant(jax_fn)
+      outputs = jax_fn(inputs)
+      if not expected_message:
+        self.assertAllClose(outputs, inputs)
+
+      checked_fn = checkify.checkify(jax_fn)
+      err, checked_outputs = checked_fn(inputs)
+      err.throw()  # Nothing to throw.
+      if not expected_message:
+        self.assertAllClose(checked_outputs, inputs)
+
+    with config.override_config("enable_checkify_for_asserts", True):
+      jax_fn = tf2jax.convert_functional(tf.function(check_numerics), inputs)
+      jax_fn = self.variant(jax_fn)
+
+      checked_fn = checkify.checkify(jax_fn)
+      err, checked_outputs = checked_fn(inputs)
+      if not expected_message:
+        self.assertAllClose(checked_outputs, inputs)
+      else:
+        with self.assertRaisesRegex(checkify.JaxRuntimeError,
+                                    f"Checking : {expected_message}"):
+          err.throw()
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_complex(self):
     reals = np.array([1.2, -2.3, 3.4], np.float32)
@@ -597,6 +649,41 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return tf.zeros(empty())
     self._test_convert(empty_static, [])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      ("exact", (10, 5), (10, 5), True),
+      ("partial0", (10, None), (10, 5), True),
+      ("partial1", (None, 5), (10, 5), True),
+      ("partial2", (None, None), (10, 5), True),
+      ("too_small", (10,), (10,), False),
+      ("too_large", (10, 5, 3), (10, 5, 3), False),
+      ("incompatible0", (20, 5), (20, 5), False),
+      ("incompatible1", (10, 7), (10, 7), False),
+      ("incompatible2", (20, None), (20, 5), False),
+      ("incompatible3", (None, 7), (10, 7), False),
+  )
+  def test_ensure_shape(self, expected_shape, example_shape, valid):
+
+    def ensure_shape(x):
+      return tf.raw_ops.EnsureShape(
+          input=x, shape=tf.TensorShape(expected_shape))
+
+    valid_inputs = np.array(
+        range(np.prod(example_shape)), dtype=np.float32).reshape(example_shape)
+    self._test_convert(ensure_shape, [valid_inputs])
+
+    jax_fn = tf2jax.convert_functional(tf.function(ensure_shape), valid_inputs)
+    jax_fn = self.variant(jax_fn)
+
+    with config.override_config("strict_shape_check", False):
+      actual_inputs = np.array(range(50), dtype=np.float32).reshape((10, 5))
+      if valid:
+        outputs = jax_fn(actual_inputs)
+        self.assertAllClose(outputs, actual_inputs)
+      else:
+        with self.assertRaisesRegex(ValueError, "Expected shape="):
+          jax_fn(actual_inputs)
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_fill(self):
     dims, value = (np.int32(2),), np.int32(3)
@@ -769,6 +856,7 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
   def test_inplace_add(self):
     np.random.seed(42)
 
+    @tf.function
     def inplace_add(x, idx, val):
       return tf.raw_ops.InplaceAdd(x=x, i=idx, v=val)
 
@@ -777,12 +865,26 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
         [2, 5],
         [[1, 2], [3, 4]],
     ]
-    self._test_convert(inplace_add, inputs)
+
+    try:
+      self._test_convert(inplace_add, inputs)
+    except tf.errors.InvalidArgumentError as e:
+      if jax.default_backend().lower() == "tpu":
+        self.assertIn("index must be Rank 1 and size 1", str(e))  # pylint: disable=g-assert-in-except
+        tpu_inputs = [
+            np.random.normal(size=[10, 2]).astype(np.float32),
+            [5],
+            [[3, 4]],
+        ]
+        self._test_convert(inplace_add, tpu_inputs)
+      else:
+        raise e
 
   @chex.variants(with_jit=True, without_jit=True)
   def test_inplace_update(self):
     np.random.seed(42)
 
+    @tf.function
     def inplace_update(x, idx, val):
       return tf.raw_ops.InplaceUpdate(x=x, i=idx, v=val)
 
@@ -793,6 +895,20 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
     ]
     self._test_convert(inplace_update, inputs)
 
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_invert_permutation(self):
+    np.random.seed(42)
+
+    def invert(x):
+      return tf.raw_ops.InvertPermutation(x=x)
+
+    inputs = np.array([2, 4, 3, 0, 1])
+    self._test_convert(invert, inputs)
+
+    def invert_static():
+      return tf.zeros(invert(inputs))
+    self._test_convert(invert_static, [])
+
   @chex.variants(with_jit=True, without_jit=True)
   @parameterized.named_parameters(
       chex.params_product(
@@ -1057,6 +1173,28 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return tf.zeros(rank(inputs))
     self._test_convert(rank_static, [])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      ("basic0", 2, 0),
+      ("basic1", -2, 1),
+      ("basic2", 0, 1),
+      ("single", (2,), (1,)),
+      ("double0", (2, -1), (0, 1)),
+      ("double1", (2, -1), (1, 0)),
+      ("dupe_axis", (2, 1), (1, 1)),
+  )
+  def test_roll(self, shift, axis):
+    inputs = np.array(range(1, 51)).reshape((10, 5))
+
+    def roll(x):
+      return tf.raw_ops.Roll(input=x, shift=shift, axis=axis)
+    self._test_convert(roll, inputs)
+
+    # Check static inputs result in static outputs.
+    def roll_static():
+      return tf.zeros(roll(inputs)[0])
+    self._test_convert(roll_static, [])
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_squeeze(self):
     inputs, dims = np.array([[[42], [47]]]), (0, 2)
@@ -1183,6 +1321,7 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
           named=True,
       ))
   def test_resize_bilinear(self, size, align_and_half):
+    np.random.seed(42)
     align, half_pixel = align_and_half
 
     def resize_bilinear(imgs):
@@ -1203,6 +1342,7 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
           named=True,
       ))
   def test_resize_bilinear_invalid(self, align_and_half):
+    np.random.seed(42)
     align, half_pixel = align_and_half
 
     def resize_bilinear(imgs):
@@ -1217,6 +1357,55 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
         ValueError, "align_corners=True and half_pixel_centers=True"):
       _ = tf2jax.convert_functional(tf.function(resize_bilinear), images)
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      chex.params_product(
+          (("upper_sample", (10, 10)), ("down_sample", (5, 5))),
+          (
+              ("not_align_and_half", (False, True)),
+          ),
+          named=True,
+      ))
+  def test_resize_nearest_neighbor(self, size, align_and_half):
+    np.random.seed(42)
+    align, half_pixel = align_and_half
+
+    def resize_nearest_neighbor(imgs):
+      return tf.raw_ops.ResizeNearestNeighbor(
+          images=imgs,
+          size=size,
+          align_corners=align,
+          half_pixel_centers=half_pixel)
+
+    images = np.random.normal(size=(4, 7, 7, 3)).astype(np.float32)
+    self._test_convert(resize_nearest_neighbor, [images])
+
+  @parameterized.named_parameters(
+      chex.params_product(
+          (
+              ("align_and_half", (True, True)),
+              ("align_and_not_half", (True, False)),
+              ("not_align_and_not_half", (False, False)),
+          ),
+          named=True,
+      ))
+  def test_resize_nearest_neighbor_invalid(self, align_and_half):
+    np.random.seed(42)
+    align, half_pixel = align_and_half
+
+    def resize_nearest_neighbor(imgs):
+      return tf.raw_ops.ResizeNearestNeighbor(
+          images=imgs,
+          size=(10, 10),
+          align_corners=align,
+          half_pixel_centers=half_pixel)
+
+    images = np.random.normal(size=(4, 7, 7, 3)).astype(np.float32)
+    with self.assertRaisesRegex(
+        ValueError, "align_corners=False and half_pixel_centers=True"):
+      _ = tf2jax.convert_functional(tf.function(resize_nearest_neighbor),
+                                    images)
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_reverse(self):
     axis = (1, 0)
@@ -1334,33 +1523,50 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return [tf.zeros(s[0]) for s in splitv(inputs)]
     self._test_convert(split_static, [])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_stateless_random_get_alg(self):
+
+    @tf.function
+    def tf_func():
+      return tf.raw_ops.StatelessRandomGetAlg()
+
+    jax_func = tf2jax.convert_functional(tf_func)
+    jax_alg = self.variant(jax_func)()
+
+    self.assertEqual(jax_alg, tf.random.Algorithm.AUTO_SELECT.value)
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_stateless_random_normal(self):
     def tf_func(seed):
       return tf.random.stateless_normal(
           shape=(16, 7), seed=seed, dtype=tf.float32)
-    self._test_convert(tf_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(tf_func, (seed,), check_shape_only=True)
 
     def raw_func(seed):
       key, counter = tf.raw_ops.StatelessRandomGetKeyCounter(seed=seed)
       return tf.raw_ops.StatelessRandomNormalV2(
           shape=(1000000, 7), key=key, counter=counter, alg=3, dtype=tf.float32)
-    self._test_convert(raw_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(raw_func, (seed,), check_shape_only=True)
 
     with self.subTest("check_statistics"):
       jax_func = tf2jax.convert_functional(
-          tf.function(raw_func), np.array([0, 42]))
+          tf.function(raw_func, jit_compile=True), seed)
       jax_func = self.variant(jax_func)
-      samples = jax_func(np.array([0, 42]))
+      samples = jax_func(seed)
       self.assertAllClose(np.mean(samples), 0.0, atol=1e-3)
       self.assertAllClose(np.std(samples), 1.0, atol=1e-3)
 
     with self.subTest("check_same_seed"):
-      same_samples = jax_func(np.array([0, 42]))
+      same_samples = jax_func(seed)
       self.assertAllClose(samples, same_samples)
 
     with self.subTest("check_diff_seed"):
-      diff_samples = jax_func(np.array([0, 47]))
+      another_seed = np.array([0, 47], dtype=np.int32)
+      diff_samples = jax_func(another_seed)
       self.assertNotAllClose(samples, diff_samples)
 
   @chex.variants(with_jit=True, without_jit=True)
@@ -1368,29 +1574,34 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
     def tf_func(seed):
       return tf.random.stateless_uniform(
           shape=(16, 7), seed=seed, dtype=tf.float32)
-    self._test_convert(tf_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(tf_func, (seed,), check_shape_only=True)
 
     def raw_func(seed):
       key, counter = tf.raw_ops.StatelessRandomGetKeyCounter(seed=seed)
       return tf.raw_ops.StatelessRandomUniformV2(
           shape=(1000000, 7), key=key, counter=counter, alg=3, dtype=tf.float32)
-    self._test_convert(raw_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(raw_func, (seed,), check_shape_only=True)
 
     with self.subTest("check_statistics"):
       jax_func = tf2jax.convert_functional(
-          tf.function(raw_func), np.array([0, 42]))
+          tf.function(raw_func, jit_compile=True), seed)
       jax_func = self.variant(jax_func)
-      samples = jax_func(np.array([0, 42]))
+      samples = jax_func(seed)
       for expected in np.linspace(0.1, 1, 10):
         actual = np.mean(samples < expected)
         self.assertAllClose(actual, expected, atol=1e-3)
 
     with self.subTest("check_same_seed"):
-      same_samples = jax_func(np.array([0, 42]))
+      same_samples = jax_func(seed)
       self.assertAllClose(samples, same_samples)
 
     with self.subTest("check_diff_seed"):
-      diff_samples = jax_func(np.array([0, 47]))
+      another_seed = np.array([0, 47], dtype=np.int32)
+      diff_samples = jax_func(another_seed)
       self.assertNotAllClose(samples, diff_samples)
 
   @chex.variants(with_jit=True, without_jit=True)
@@ -1398,7 +1609,9 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
     def tf_func(seed):
       return tf.random.stateless_uniform(
           shape=(16, 7), seed=seed, minval=0, maxval=10, dtype=tf.int32)
-    self._test_convert(tf_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(tf_func, (seed,), check_shape_only=True)
 
     def raw_func(seed):
       key, counter = tf.raw_ops.StatelessRandomGetKeyCounter(seed=seed)
@@ -1409,23 +1622,26 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
           alg=3,
           minval=0,
           maxval=10)
-    self._test_convert(raw_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(raw_func, (seed,), check_shape_only=True)
 
     with self.subTest("check_statistics"):
       jax_func = tf2jax.convert_functional(
-          tf.function(raw_func), np.array([0, 42]))
+          tf.function(raw_func, jit_compile=True), seed)
       jax_func = self.variant(jax_func)
-      samples = jax_func(np.array([0, 42]))
+      samples = jax_func(seed)
       for val in range(0, 10):
         actual = np.mean(samples == val)
         self.assertAllClose(actual, 0.1, atol=1e-3)
 
     with self.subTest("check_same_seed"):
-      same_samples = jax_func(np.array([0, 42]))
+      same_samples = jax_func(seed)
       self.assertAllClose(samples, same_samples)
 
     with self.subTest("check_diff_seed"):
-      diff_samples = jax_func(np.array([0, 47]))
+      another_seed = np.array([0, 47], dtype=np.int32)
+      diff_samples = jax_func(another_seed)
       self.assertNotAllClose(samples, diff_samples)
 
   @chex.variants(with_jit=True, without_jit=True)
@@ -1434,13 +1650,15 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       key, counter = tf.raw_ops.StatelessRandomGetKeyCounter(seed=seed)
       return tf.raw_ops.StatelessRandomUniformFullIntV2(
           shape=(1000000, 7), key=key, counter=counter, alg=3, dtype=tf.int32)
-    self._test_convert(raw_func, (np.array([0, 42]),), check_shape_only=True)
+
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(raw_func, (seed,), check_shape_only=True)
 
     with self.subTest("check_statistics"):
       jax_func = tf2jax.convert_functional(
-          tf.function(raw_func), np.array([0, 42]))
+          tf.function(raw_func, jit_compile=True), seed)
       jax_func = self.variant(jax_func)
-      samples = jax_func(np.array([0, 42]))
+      samples = jax_func(seed)
       int32_min = np.iinfo(np.int32).min
       int32_max = np.iinfo(np.int32).max
       for val in range(int32_min, int32_max, 200_000_000):
@@ -1449,29 +1667,29 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
         self.assertAllClose(actual, expected, atol=1e-3)
 
     with self.subTest("check_same_seed"):
-      same_samples = jax_func(np.array([0, 42]))
+      same_samples = jax_func(seed)
       self.assertAllClose(samples, same_samples)
 
     with self.subTest("check_diff_seed"):
-      diff_samples = jax_func(np.array([0, 47]))
+      another_seed = np.array([0, 47], dtype=np.int32)
+      diff_samples = jax_func(another_seed)
       self.assertNotAllClose(samples, diff_samples)
 
   @chex.variants(with_jit=True, without_jit=True)
   def test_stateless_multinomial(self):
-    inputs = np.array([np.arange(5), np.arange(5, 0, -1)], dtype=np.float32)
-
     def raw_func(logits, seed):
       return tf.raw_ops.StatelessMultinomial(
           logits=logits, num_samples=1000000, seed=seed, output_dtype=tf.int32)
 
-    self._test_convert(
-        raw_func, (inputs, np.array([0, 42])), check_shape_only=True)
+    inputs = np.array([np.arange(5), np.arange(5, 0, -1)], dtype=np.float32)
+    seed = np.array([0, 42], dtype=np.int32)
+    self._test_convert(raw_func, (inputs, seed), check_shape_only=True)
 
     with self.subTest("check_statistics"):
       jax_func = tf2jax.convert_functional(
-          tf.function(raw_func), inputs, np.array([0, 42]))
+          tf.function(raw_func, jit_compile=True), inputs, seed)
       jax_func = self.variant(jax_func)
-      samples = jax_func(inputs, np.array([0, 42]))
+      samples = jax_func(inputs, seed)
       for batch in range(inputs.shape[0]):
         probs = jax.nn.softmax(inputs[batch])
         samps = samples[batch]
@@ -1479,11 +1697,12 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
           self.assertAllClose(probs[idx], np.mean(samps == idx), atol=1e-3)
 
     with self.subTest("check_same_seed"):
-      same_samples = jax_func(inputs, np.array([0, 42]))
+      same_samples = jax_func(inputs, seed)
       self.assertAllClose(samples, same_samples)
 
     with self.subTest("check_diff_seed"):
-      diff_samples = jax_func(inputs, np.array([0, 47]))
+      another_seed = np.array([0, 47], dtype=np.int32)
+      diff_samples = jax_func(inputs, another_seed)
       self.assertNotAllClose(samples, diff_samples)
 
   @chex.variants(with_jit=True, without_jit=True)
@@ -1498,6 +1717,7 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       ("shrink0", lambda x: x[..., 2, 3]),
       ("shrink1", lambda x: x[2:, 3, :, 5]),
       ("mixed0", lambda x: x[..., 4:1:-1, tf.newaxis, 3]),
+      ("zero_slice", lambda x: x[:, :0, ...]),
   )
   def test_strided_slice(self, slice_fn):
     inputs = np.linspace(0., 1., 4 * 5 * 6 * 7).astype(np.float32)
@@ -1534,6 +1754,99 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       )
     self._test_convert(switch_fn, [np.array(1, dtype=np.int32)])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      ("defined", (3, 2)),
+      ("partial0", (-1, 2)),
+      ("partial1", (3, -1)),
+      ("undefined", -1)
+  )
+  def test_tensor_list_get_item(self, init_shape):
+    @tf.function(jit_compile=False)
+    def tensor_list_fn():
+      handle = tf.raw_ops.TensorListReserve(
+          element_shape=init_shape, num_elements=3, element_dtype=tf.int32)
+      return tf.raw_ops.TensorListGetItem(
+          input_handle=handle,
+          index=1,
+          element_shape=(3, 2),
+          element_dtype=tf.int32)
+
+    self._test_convert(tensor_list_fn, [])
+
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      ("defined", dict(fn_output_signature=tf.TensorSpec((5, 4), tf.int32))),
+      ("undefined", {})
+  )
+  def test_tensor_list(self, output_signature_kwargs):
+    def tensor_list_fn():
+      return tf.map_fn(
+          fn=lambda t: tf.ones((5, 4), dtype=tf.int32) * t,
+          elems=tf.constant([3, 5, 2]),
+          **output_signature_kwargs,
+      )
+
+    self._test_convert(tensor_list_fn, [])
+
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      chex.params_product(
+          (
+              (
+                  "1D",
+                  np.ones([8], dtype=np.float32),
+                  np.array([[4], [3], [1], [7]], dtype=np.int32),
+                  np.array([9, 10, 11, 12], dtype=np.float32),
+              ),
+              (
+                  "2D_scalar",
+                  np.ones([8, 2], dtype=np.float32),
+                  np.array([[4, 0], [3, 1], [1, 0], [7, 1]], dtype=np.int32),
+                  np.array([9, 10, 11, 12], dtype=np.float32),
+              ),
+              (
+                  "2D_slice",
+                  np.ones([8, 2], dtype=np.float32),
+                  np.array([[4], [3], [1], [7]], dtype=np.int32),
+                  np.array([[9, 90], [10, 100], [11, 110], [12, 120]],
+                           dtype=np.float32),
+              ),
+              (
+                  "3D_scalar",
+                  np.ones([8, 3, 2], dtype=np.float32),
+                  np.array([[4, 0, 0], [3, 1, 1], [1, 2, 0], [7, 0, 1]],
+                           dtype=np.int32),
+                  np.array([9, 10, 11, 12], dtype=np.float32),
+              ),
+              (
+                  "3D_slice",
+                  np.ones([8, 3, 2], dtype=np.float32),
+                  np.array([[4, 0], [3, 1], [1, 2], [7, 0]], dtype=np.int32),
+                  np.array([[9, 90], [10, 100], [11, 110], [12, 120]],
+                           dtype=np.float32),
+              ),
+              (
+                  "3D_block",
+                  np.ones([8, 3, 2], dtype=np.float32),
+                  np.array([[4], [3], [1], [7]], dtype=np.int32),
+                  np.array([
+                      [[9, 90], [91, 92], [93, 94]],
+                      [[10, 100], [101, 102], [103, 104]],
+                      [[11, 110], [111, 112], [113, 114]],
+                      [[12, 120], [121, 122], [123, 124]],
+                  ],
+                           dtype=np.float32),
+              ),
+          ),
+          named=True,
+      ))
+  def test_tensor_scatter_update(self, tensor, indices, updates):
+    def scatter(x, inds, ups):
+      return tf.raw_ops.TensorScatterUpdate(tensor=x, indices=inds, updates=ups)
+
+    self._test_convert(scatter, [tensor, indices, updates])
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_unpack(self):
     inputs = np.array([[1, 2], [3, 4], [5, 6]])
@@ -1547,6 +1860,22 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return [tf.zeros(s) for s in unpack(inputs)]
     self._test_convert(unpack_static, [])
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.parameters("UnsortedSegmentSum", "UnsortedSegmentMax",
+                            "UnsortedSegmentMin", "UnsortedSegmentProd")
+  def test_unsorted_segment(self, op_name):
+    def segment_reduce(x, ids):
+      return getattr(tf.raw_ops, op_name)(
+          data=x, segment_ids=ids, num_segments=2)
+
+    data = np.array([5, 1, 7, 2, 3, 4], np.float32)
+    segment_ids = np.array([0, 0, 1, 1, 0, 1], np.int32)
+    self._test_convert(segment_reduce, [data, segment_ids])
+
+    data = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]], np.float32)
+    segment_ids = np.array([0, 1, 0], np.int32)
+    self._test_convert(segment_reduce, [data, segment_ids])
+
   @chex.variants(without_jit=True)
   def test_where(self):
     inputs = [np.array([True, False])]
@@ -1624,7 +1953,17 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
     @tf.function
     def while_loop(x):
       return tf.while_loop(cond, body, [x, step])
-    self._test_convert(while_loop, np.zeros(output_size, np.float32))
+
+    if jax.default_backend().lower() == "tpu":
+      tpu_context = self.assertRaisesRegex(
+          tf.errors.InvalidArgumentError,
+          ("Input 0 to node `while/ones` with op Fill must be a compile-time "
+           "constant."))
+    else:
+      tpu_context = contextlib.nullcontext()
+
+    with tpu_context:
+      self._test_convert(while_loop, np.zeros(output_size, np.float32))
 
   @chex.variants(with_jit=True, without_jit=True)
   def test_assign_side_effect(self):
@@ -1649,6 +1988,59 @@ class OpsTest(tf.test.TestCase, parameterized.TestCase):
       return tf.raw_ops.TopKV2(input=x, k=k, sorted=True)
     self._test_convert(top_k, inputs)
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.parameters((2., 2.), (2., 0.), (0., 0.), (0., 2.))
+  def test_div_no_nan(self, x, y):
+
+    @tf.function
+    def div_no_nan(inputs):
+      x, y = inputs
+      return tf.math.divide_no_nan(x=x, y=y)
+
+    x = np.array(x)
+    y = np.array(y)
+    tf_x = tf.convert_to_tensor(x)
+    tf_y = tf.convert_to_tensor(y)
+
+    with tf.GradientTape() as g:
+      g.watch(tf_x)
+      g.watch(tf_y)
+      output = div_no_nan([tf_x, tf_y])
+    tf_gradient = g.gradient(output, [tf_x, tf_y])
+
+    jax_func = tf2jax.convert_functional(div_no_nan, [x, y])
+    jax_func = self.variant(jax_func)
+    jax_gradient = jax.grad(jax_func)([x, y])
+
+    with self.subTest("check_forward_pass"):
+      self.assertAllClose(jax_func([x, y]), np.asarray(output))
+    with self.subTest("check_backward_pass"):
+      self.assertAllClose(jax_gradient, np.asarray(tf_gradient))
+
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_angle(self):
+    inputs = np.array([-2.25 + 4.75j, 3.25 + 5.75j], dtype=np.csingle)
+
+    def angle(x):
+      return tf.raw_ops.Angle(input=x)
+    self._test_convert(angle, inputs)
+
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_rfft(self):
+    inputs = np.array([2.25, 3.25] * 2, dtype=np.single)
+
+    def rfft(x):
+      return tf.raw_ops.RFFT(input=x, fft_length=[len(x)])
+    self._test_convert(rfft, inputs)
+
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_irfft(self):
+    inputs = np.array([-2.25 + 4.75j, 3.25 + 5.75j] * 2, dtype=np.csingle)
+
+    def irfft(x):
+      return tf.raw_ops.IRFFT(input=x, fft_length=[len(x)])
+    self._test_convert(irfft, inputs)
+
 
 if __name__ == "__main__":
   tf.test.main()
