@@ -26,7 +26,9 @@ import jax.numpy as jnp
 import numpy as np
 import tensorflow as tf
 from tf2jax._src import config
+from tf2jax._src import test_util
 from tf2jax._src import tf2jax
+from tf2jax._src.experimental import ops as experimental_ops  # pylint: disable=unused-import
 import tree
 
 from tensorflow.compiler.xla import xla_data_pb2  # pytype: disable=import-error
@@ -41,17 +43,26 @@ def _compute_gradients(func, *inputs):
   return jax.grad(lambda *args: jnp.sum(fn(*args)))(*inputs)
 
 
-class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
+class Jax2TfTest(test_util.TestCase):
 
   def _test_convert(
       self,
       jax_func,
       inputs,
+      *,
       with_grad,
       enable_xla,
       with_custom_grad=False,
       grad_tols=None,
   ):
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      if not enable_xla:
+        self.skipTest("native_lowering does not support enable_xla=False.")
+      if with_grad and not with_custom_grad:
+        self.skipTest(
+            "native_lowering does not support differentiation without custom "
+            "gradient.")
+
     grad_tols = grad_tols or {}
     def assert_grad_all_close(*args):
       return self.assertAllClose(*args, **grad_tols)
@@ -65,13 +76,14 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     # Jax -> TF
     tf_func = jax2tf.convert(jax_func, with_gradient=with_grad,
                              enable_xla=enable_xla)
+    tf_func = tf.function(tf_func, jit_compile=True, autograph=False)
     tf_outputs = tf_func(*inputs)
     jax.tree_map(self.assertAllClose, jax_outputs, tf_outputs)
 
     # Jax -> TF -> Jax
     with config.override_config("convert_custom_gradient", with_custom_grad):
       rejax_func = tf2jax.convert_functional(
-          tf.function(tf_func), *tree.map_structure(np.zeros_like, inputs))
+          tf_func, *tree.map_structure(np.zeros_like, inputs))
     rejax_func = self.variant(rejax_func)
     rejax_outputs = rejax_func(*inputs)
     jax.tree_map(self.assertAllClose, rejax_outputs, tf_outputs)
@@ -81,8 +93,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
 
     # Jax -> TF -> SavedModel -> TF
     model = tf.Module()
-    model.f = tf.function(tf_func)
-    model.f(*tree.map_structure(tf.zeros_like, inputs))  # Dummy call.
+    model.f = tf_func
     tmp_dir = self.create_tempdir()
     tf.saved_model.save(model, tmp_dir.full_path)
     del model
@@ -115,7 +126,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     def forward(x):
       return jnp.sin(jnp.cos(x))
 
-    inputs = np.random.normal((3, 2))
+    inputs = np.random.normal((3, 2)).astype(np.float32)
     self._test_convert(
         forward, [inputs],
         with_grad=with_grad,
@@ -431,6 +442,61 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     self._test_convert(
         forward, [inputs], with_grad=with_grad, enable_xla=enable_xla)
 
+  @chex.variants(with_jit=True)
+  @parameterized.named_parameters(
+      chex.params_product(
+          (("without_gradient", False), ("with_gradient", True)),
+          (("enable_xla", True),),
+          (
+              ("min", jax.lax.cummin),
+              ("max", jax.lax.cummax),
+              ("sum", jax.lax.cumsum),
+              ("prod", jax.lax.cumprod),
+          ),
+          (
+              ("dim0", 0),
+              ("dim1", 1),
+          ),
+          (
+              ("reverse", True),
+              ("not_reverse", False),
+          ),
+          (
+              ("heuristic", True),
+              ("no_heuristic", False),
+          ),
+          named=True,
+      ))
+  def test_cumulative_reduction(self, with_grad, enable_xla, reducer, axis,
+                                reverse, use_heuristic):
+    if (with_grad and not use_heuristic and
+        jax.default_backend().lower() == "tpu"):
+      self.skipTest("Gradient of reduce-window not always supported on TPU")
+
+    np.random.seed(42)
+
+    def forward(x):
+      return reducer(x, axis=axis, reverse=reverse)
+
+    inputs = np.random.normal(size=(4, 3))
+
+    with config.override_config("infer_cumulative_reduction_from_jax2tf",
+                                use_heuristic):
+      roundtrip_forward = tf2jax.convert_functional(
+          tf.function(jax2tf.convert(forward), autograph=False), inputs)
+      roundtrip_jaxpr = jax.make_jaxpr(roundtrip_forward)(inputs)
+      if (use_heuristic and
+          not jax.config.jax2tf_default_experimental_native_lowering):
+        self.assertNotIn("reduce_window", roundtrip_jaxpr.pretty_print())
+
+      if (with_grad and enable_xla and reducer is jax.lax.cumprod and
+          not use_heuristic):
+        self.skipTest("No differentiation rule for `reduce_window` with "
+                      "`jax.lax.cumprod`.")
+
+      self._test_convert(
+          forward, [inputs], with_grad=with_grad, enable_xla=enable_xla)
+
   @chex.variants(with_jit=True)
   @parameterized.named_parameters(
       chex.params_product(
@@ -519,8 +585,16 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
 
     inputs = np.random.normal(size=(8, 5, 5, 3))
     jax_fn = jax.jacrev(forward)
-    self._test_convert(
-        jax_fn, [inputs], with_grad=with_grad, enable_xla=enable_xla)
+
+    try:
+      self._test_convert(
+          jax_fn, [inputs], with_grad=with_grad, enable_xla=enable_xla)
+    except tf.errors.InvalidArgumentError as e:
+      if jax.default_backend().lower() == "tpu":
+        # Can fail on older TPUs.
+        self.assertIn("XLA can't use dynamic begin values for slice", str(e))  # pylint: disable=g-assert-in-except
+      else:
+        raise e
 
   @chex.variants(with_jit=True)
   @parameterized.named_parameters(
@@ -552,12 +626,20 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
           (("without_gradient", False), ("with_gradient", True)),
           (("disable_xla", False), ("enable_xla", True)),
           named=True,
-      ))
+      )
+  )
   def test_polymorphic_shape(self, with_grad, enable_xla):
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      if not enable_xla:
+        self.skipTest("native_lowering does not support enable_xla=False.")
+      if enable_xla:
+        # TODO(b/267480040): Clean this up accordingly.
+        self.skipTest("re-enable after shape refinement is finished.")
+
     inputs = np.array(range(36), dtype=np.float32).reshape(9, 4)
 
     # TF
-    @tf.function
+    @tf.function(autograph=False)
     def forward(x):
       shape = tf.shape(x)
       # Manipulate polymoprhic shape.
@@ -569,8 +651,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     tf_outputs = forward(inputs)
 
     # TF -> JAX
-    jax_func = tf2jax.convert_functional(
-        tf.function(forward), tf.TensorSpec((None, 4)))
+    jax_func = tf2jax.convert_functional(forward, tf.TensorSpec((None, 4)))
     jax_func = self.variant(jax_func)
     jax_outputs = jax_func(inputs)
     self.assertAllClose(tf_outputs, jax_outputs)
@@ -581,7 +662,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
         polymorphic_shapes=["(b, _)"],
         with_gradient=with_grad,
         enable_xla=enable_xla)
-    new_tf_forward = tf.function(new_tf_forward)
+    new_tf_forward = tf.function(new_tf_forward, autograph=False)
     concrete_new_tf_forward = new_tf_forward.get_concrete_function(
         tf.TensorSpec(shape=(None, 4)))
     self.assertEqual(concrete_new_tf_forward.structured_outputs.shape.as_list(),
@@ -597,6 +678,10 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
           named=True,
       ))
   def test_custom_gradient(self, with_grad, enable_xla):
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      if not enable_xla:
+        self.skipTest("native_lowering does not support enable_xla=False.")
+
     inputs = np.array(range(6), dtype=np.float32).reshape(3, 2)
 
     # JAX
@@ -614,7 +699,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     # JAX -> TF
     tf_forward = jax2tf.convert(
         forward, with_gradient=with_grad, enable_xla=enable_xla)
-    tf_forward = tf.function(tf_forward)
+    tf_forward = tf.function(tf_forward, autograph=False)
 
     # JAX -> TF -> JAX
     with config.override_config("convert_custom_gradient", True):
@@ -652,6 +737,10 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
           named=True,
       ))
   def test_custom_gradient_nested(self, with_grad, enable_xla):
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      if not enable_xla:
+        self.skipTest("native_lowering does not support enable_xla=False.")
+
     inputs = np.array(range(6), dtype=np.float32).reshape(3, 2)
 
     # JAX
@@ -670,13 +759,13 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     # JAX -> TF
     tf_fn = jax2tf.convert(
         forward, with_gradient=with_grad, enable_xla=enable_xla)
-    tf_fn = tf.function(tf_fn)
+    tf_fn = tf.function(tf_fn, autograph=False)
 
     # JAX -> TF -> CALL_TF -> TF.
     # This creates dependencies between custom gradients.
     call_tf_fn = jax2tf.call_tf(tf_fn)
     tf_fn_too = jax2tf.convert(call_tf_fn, with_gradient=with_grad)
-    tf_fn_too = tf.function(tf_fn_too)
+    tf_fn_too = tf.function(tf_fn_too, autograph=False)
 
     # JAX -> TF -> CALL_TF -> TF -> JAX
     with config.override_config("convert_custom_gradient", True):
@@ -710,6 +799,10 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
           named=True,
       ))
   def test_empty_return(self, enable_xla):
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      if not enable_xla:
+        self.skipTest("native_lowering does not support enable_xla=False.")
+
     np.random.seed(42)
 
     def forward(x):
@@ -727,7 +820,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
     # JAX -> TF
     tf_forward = jax2tf.convert(
         forward, with_gradient=True, enable_xla=enable_xla)
-    tf_forward = tf.function(tf_forward)
+    tf_forward = tf.function(tf_forward, autograph=False)
 
     # JAX -> TF -> JAX
     with config.override_config("convert_custom_gradient", True):
@@ -761,7 +854,7 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
 
   @chex.variants(with_jit=True, without_jit=True)
   def test_jax2tf_nesting(self):
-    @tf.function
+    @tf.function(autograph=False)
     def tf_fn(x):
       return tf.cos(x)
 
@@ -770,10 +863,73 @@ class Jax2TfTest(parameterized.TestCase, tf.test.TestCase):
       return jnp.sin(self.variant(jax_fn)(x))
 
     tf2jax2tf_fn = jax2tf.convert(tf2jax_fn)
+    tf2jax2tf_fn = tf.function(tf2jax2tf_fn, autograph=False)
 
     inputs = np.linspace(-1., 1., 6, dtype=np.float32).reshape((2, 3))
     self.assertAllClose(tf.sin(tf_fn(inputs)), tf2jax2tf_fn(inputs))
 
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      chex.params_product(
+          (("without_gradient", False), ("with_gradient", True)),
+          named=True,
+      ))
+  def test_remat(self, with_gradient):
+    def fn(x):
+      return jnp.sin(jnp.sin(x))
+    remat_fn = jax.checkpoint(fn)
+
+    inputs = (
+        np.linspace(0, 1, 10 * 5, dtype=np.float32).reshape(10, 5),
+    )
+    self._test_convert(
+        remat_fn,
+        inputs,
+        with_grad=with_gradient,
+        enable_xla=True,
+        with_custom_grad=True)
+
+    if jax.config.jax2tf_default_experimental_native_lowering:
+      self.skipTest("Skip remat jaxpr test with native_lowering.")
+
+    # Check jaxpr.
+    tf_fn = tf.function(
+        jax2tf.convert(remat_fn, with_gradient=True, enable_xla=True),
+        autograph=False)
+    jax_fn = tf2jax.convert_functional(tf_fn, tf.TensorSpec((10, 5),
+                                                            tf.float32))
+    jax_fn = self.variant(jax_fn)
+    out_jaxpr = jax.make_jaxpr(jax_fn)(*inputs)
+    self.assertNotRegex(str(out_jaxpr), "remat")
+    grad_jaxpr = jax.make_jaxpr(jax.grad(lambda x: jnp.sum(jax_fn(x))))(*inputs)
+    self.assertRegex(str(grad_jaxpr), "optimization_barrier")
+
+  @chex.variants(with_jit=True, without_jit=True)
+  @parameterized.named_parameters(
+      chex.params_product(
+          (("without_gradient", False),),
+          (("enable_xla", True),),
+          (("without_custom_gradient", False), ("with_custom_gradient", True)),
+          named=True,
+      ))
+  def test_reduce_precision(self, with_grad, enable_xla, with_custom_grad):
+    if jax.__version_info__ <= (0, 4, 4):
+      self.skipTest("jax.lax.reduce_precision is only supported from 0.4.4 and "
+                    f"onward, found {jax.__version__}.")
+
+    np.random.seed(42)
+
+    def forward(x):
+      info = jnp.finfo(jnp.bfloat16)
+      return jax.lax.reduce_precision(x, info.nexp, info.nmant)
+
+    inputs = np.random.normal((3, 2)).astype(np.float32)
+    self._test_convert(
+        forward, [inputs],
+        with_grad=with_grad,
+        enable_xla=enable_xla,
+        with_custom_grad=with_custom_grad)
+
 
 if __name__ == "__main__":
   absltest.main()
