@@ -231,7 +231,7 @@ class _OpNode:
 
     # Return updated variables.
     if self.op in _TF_ASSIGN_OPS:
-      # TODO(shaobohou) This assumes the assign op returns the updated value.
+      # This assumes the assign op returns the updated value.
       updated_params = {named_args[0][0]: outputs}
     else:
       updated_params = {}
@@ -355,24 +355,35 @@ def convert_functional_from_restored(tf_func: Any) -> AnnotatedFunction:
 def convert_from_restored(
     tf_func) -> Tuple[AnnotatedFunction, Mapping[str, Variable]]:
   """Converts a RestoredFunction (from a SaveModel) if it is unambiguous."""
-  concrete_functions = getattr(tf_func, "concrete_functions", ())
-  if len(concrete_functions) != 1:
-    err_message = (
-        f"Found {len(concrete_functions)} concrete functions, use "
-        "tf2jax.convert or tf2jax.convert_functional with *args and **kwargs "
-        "to select one of the options above.")
-    saved_model_err_message = ""
-    try:
-      # Deliberately trigger pretty error message from SavedModel.
-      tf_func(None)
-    except ValueError as e:
-      saved_model_err_message = str(e)
-    raise ValueError(saved_model_err_message + "\n\n" + err_message)
+  if tf_func.input_signature is None:
+    concrete_functions = getattr(tf_func, "concrete_functions", ())
+    if len(concrete_functions) != 1:
+      err_message = (
+          f"Found {len(concrete_functions)} concrete functions, use "
+          "tf2jax.convert or tf2jax.convert_functional with *args and **kwargs "
+          "to select one of the options above.")
+      saved_model_err_message = ""
+      try:
+        # Deliberately trigger pretty error message from SavedModel.
+        tf_func(None)
+      except ValueError as e:
+        saved_model_err_message = str(e)
+      raise ValueError(saved_model_err_message + "\n\n" + err_message)
+
+    args, kwargs = concrete_functions[0].structured_input_signature
+  else:
+    args = tf_func.input_signature
+    kwargs = {}
 
-  args, kwargs = concrete_functions[0].structured_input_signature
   return convert(tf_func, *args, **kwargs)
 
 
+def _is_tensorspec_like(v: Any) -> bool:
+  return (isinstance(v, tf.TensorSpec) or (
+      type(v).__module__.endswith("tensorflow.python.framework.tensor_spec") and
+      type(v).__name__ == "VariableSpec"))
+
+
 def _fix_tfhub_specs(
     structured_specs,
     flat_tensors,
@@ -382,7 +393,7 @@ def _fix_tfhub_specs(
   flat_specs = tree.flatten(structured_specs)
   tensor_count = 0
   for idx, val in enumerate(flat_specs):
-    if isinstance(val, tf.TensorSpec):
+    if _is_tensorspec_like(val):
       flat_specs[idx] = tf.TensorSpec(
           val.shape, val.dtype, name=flat_tensors[tensor_count].op.name)
       tensor_count += 1
@@ -461,8 +472,8 @@ def convert(
   flat_outputs = tree.flatten(concrete_func.outputs)
   structured_outputs = tree.map_structure(maybe_tensor_to_spec,
                                           concrete_func.structured_outputs)
-  # TODO(shaobohou) We do not check the number of output tensors because they do
-  # not match for custom gradients functions.
+  # We do not check the number of output tensors because they do not match for
+  # custom gradients functions.
   structured_outputs = _fix_tfhub_specs(structured_outputs, flat_outputs)
   try:
     fullargspec = tf_func.function_spec.fullargspec
@@ -478,7 +489,7 @@ def convert(
       raise ValueError("If function_spec is None then only keyword arguments "
                        f"are expectd, found args={exp_args} in structure.")
     parameters = tuple([
-        # TODO(shaobohou) Remove temporary fix for TF-Hub.
+        # TODO(b/266552275) Remove temporary fix for TF-Hub.
         inspect.Parameter(
             k.replace("$", "___"), kind=inspect.Parameter.KEYWORD_ONLY)
         for k in tree.flatten(exp_kwargs.keys())
@@ -753,20 +764,23 @@ def _infer_relu_from_jax2tf(nodes):
   found_jax2tf = any(["jax2tf_out" in n.name for n in nodes])
   node_map = {n.name: n for n in nodes}
   for node in nodes:
-    if node.op == "Maximum":
-      cast_arg = node_map[node.inputs[1].op_name]
-      if cast_arg.op == "Cast":
-        const_arg = node_map[cast_arg.inputs[0].op_name]
-        if const_arg.op == "Const" and const_arg((), rng=None)[0].tolist() == 0:
-          # Replace the Maximum op with a Relu op, but keep the node name.
-          # The Cast and Const ops may now be redundant but are kept anyway.
-          node.op = "Relu"
-          node.inputs = node.inputs[:1]
-          node.jax_func = (
-              ops.get_parser("Relu")(_NodeDef("Relu", node.name, (), {})))
-          if not found_jax2tf:
-            logging.warning("Replaced max(x, 0) with jax.nn.relu but did not "
-                            "find jax2tf_out.")
+    if node.op == "Maximum" and "jax2tf" in node.name and "_relu_" in node.name:
+      cast_or_const_arg = node_map[node.inputs[1].op_name]
+      if cast_or_const_arg.op == "Cast":
+        const_arg = node_map[cast_or_const_arg.inputs[0].op_name]
+      else:
+        const_arg = cast_or_const_arg
+
+      if const_arg.op == "Const" and const_arg((), rng=None)[0].tolist() == 0:
+        # Replace the Maximum op with a Relu op, but keep the node name.
+        # The Cast and Const ops may now be redundant but are kept anyway.
+        node.op = "Relu"
+        node.inputs = node.inputs[:1]
+        node.jax_func = (
+            ops.get_parser("Relu")(_NodeDef("Relu", node.name, (), {})))
+        if not found_jax2tf:
+          logging.warning("Replaced max(x, 0) with jax.nn.relu but did not "
+                          "find jax2tf_out.")
 
 
 _FunctionDef = Any
@@ -863,7 +877,7 @@ def _convert(
   input_path_to_specs = [(p, v)
                          for p, v in tree.flatten_with_path(structured_inputs)
                          if v is not _UNUSED_INPUT]
-  # TODO(shaobohou) Remove temporary fix for TF-Hub.
+  # TODO(b/266552275) Remove temporary fix for TF-Hub.
   replace_fn = (lambda k: k.replace("$", "___") if isinstance(k, str) else k)
   input_path_to_specs = [
       (tree.map_structure(replace_fn, p), v) for p, v in input_path_to_specs
@@ -956,7 +970,6 @@ def _convert(
     for _, subgraph in subgraphs.items():
       nodes = subgraph.rewrite(nodes)
 
-  # TODO(shaobohou) restore some sort of signature
   def jax_func(
       params: Mapping[str, jnp.ndarray],
       *func_args,
@@ -1138,7 +1151,7 @@ def _filter_nodes(
       yield node, graph
   if hasattr(graph_def, "library"):
     for func in graph_def.library.function:
-      # TODO(shaobohou) Use the public API once it is available.
+      # TODO(b/266553384) Use the public API once it is available.
       library_fn = graph._functions[func.signature.name]  # pylint: disable=protected-access
       for node in func.node_def:
         if predicate(node):
