@@ -14,6 +14,8 @@
 # ==============================================================================
 """Tests tf2jax."""
 
+import inspect
+
 from absl.testing import parameterized
 
 import chex
@@ -24,10 +26,41 @@ import numpy as np
 
 import sonnet as snt
 import tensorflow as tf
+from tf2jax._src import config
 from tf2jax._src import tf2jax
 import tree
 
 
+class TestDense(tf.Module):
+
+  def __init__(self, input_dim, output_size, name=None):
+    super().__init__(name=name)
+    self.w = tf.Variable(tf.random.normal([input_dim, output_size]), name="w")
+    self.b = tf.Variable(tf.zeros([output_size]), name="b")
+
+  def __call__(self, x):
+    y = tf.matmul(x, self.w) + self.b
+    return tf.nn.relu(y)
+
+
+class TestMLP(tf.Module):
+  """Variables for different dense layers will not have unique names."""
+
+  def __init__(self, input_size, sizes, name=None):
+    super().__init__(name=name)
+    self.layers = []
+    with self.name_scope:
+      for size in sizes:
+        self.layers.append(TestDense(input_dim=input_size, output_size=size))
+        input_size = size
+
+  @tf.Module.with_name_scope
+  def __call__(self, x):
+    for layer in self.layers:
+      x = layer(x)
+    return x
+
+
 class ModelsTest(tf.test.TestCase, parameterized.TestCase):
 
   def _test_convert(self, tf_func, inputs):
@@ -133,7 +166,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
 
       test_inputs = np.ones([20, 5], dtype=np.float32)
       expected_outputs = tf_func(test_inputs)
-      with tf2jax.override_config("strict_shape_check", False):
+      with config.override_config("strict_shape_check", False):
         actual_outputs, _ = jax_func({}, test_inputs)
       self.assertAllClose(expected_outputs, actual_outputs)
 
@@ -150,7 +183,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
 
     test_inputs = np.ones([20, 7, 2], dtype=np.float32)
     expected_outputs = tf_func(test_inputs)
-    with tf2jax.override_config("strict_shape_check", False):
+    with config.override_config("strict_shape_check", False):
       actual_outputs, _ = jax_func({}, test_inputs)
     self.assertAllClose(expected_outputs, actual_outputs)
 
@@ -164,7 +197,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
 
     test_inputs = np.ones([20, 5], dtype=np.float32)
     expected_outputs = tf_func(test_inputs)
-    with tf2jax.override_config("strict_shape_check", False):
+    with config.override_config("strict_shape_check", False):
       actual_outputs = jax_func(test_inputs)
     self.assertAllClose(expected_outputs, actual_outputs)
 
@@ -200,30 +233,30 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
 
     with self.subTest("valid_input_shape"):
       expected_outputs = tf_func(orig_inputs)
-      with tf2jax.override_config("strict_shape_check", True):
+      with config.override_config("strict_shape_check", True):
         actual_outputs = jax_func(orig_inputs)
         self.assertAllClose(expected_outputs, actual_outputs)
-      with tf2jax.override_config("strict_dtype_check", True):
+      with config.override_config("strict_dtype_check", True):
         actual_outputs = jax_func(orig_inputs)
         self.assertAllClose(expected_outputs, actual_outputs)
 
     test_inputs = np.ones([10, 5, 2], dtype=np.float32)
     with self.subTest("invalid_input_shape"):
       expected_outputs = tf_func(test_inputs)
-      with tf2jax.override_config("strict_shape_check", True):
+      with config.override_config("strict_shape_check", True):
         with self.assertRaisesRegex(ValueError, "incompatible input shape"):
           actual_outputs = jax_func(test_inputs)
-      with tf2jax.override_config("strict_shape_check", False):
+      with config.override_config("strict_shape_check", False):
         actual_outputs = jax_func(test_inputs)
         self.assertNotAllClose(expected_outputs, actual_outputs)
 
     test_inputs = np.ones([10, 5], dtype=np.int32)
     with self.subTest("invalid_input_dtype"):
       expected_outputs = tf_func(test_inputs)
-      with tf2jax.override_config("strict_dtype_check", True):
+      with config.override_config("strict_dtype_check", True):
         with self.assertRaisesRegex(ValueError, "incompatible input dtype"):
           actual_outputs = jax_func(test_inputs)
-      with tf2jax.override_config("strict_dtype_check", False):
+      with config.override_config("strict_dtype_check", False):
         actual_outputs = jax_func(test_inputs)
         self.assertAllClose(expected_outputs, actual_outputs)
 
@@ -246,7 +279,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
                          "force_const_float64_to_bfloat16")
 
     # This is the default.
-    with tf2jax.override_config(dtype_config_name, False):
+    with config.override_config(dtype_config_name, False):
       jax_func = tf2jax.convert_functional(tf_func, np_inputs)
     jax_func = self.variant(jax_func)
     orig_jax_outputs = jax_func(np_inputs)
@@ -255,7 +288,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
         jnp.array(orig_jax_outputs).dtype)
     self.assertAllClose(tf_outputs, orig_jax_outputs)
 
-    with tf2jax.override_config(dtype_config_name, True):
+    with config.override_config(dtype_config_name, True):
       jax_func = tf2jax.convert_functional(tf_func, np_inputs)
     jax_func = self.variant(jax_func)
     forced_jax_outputs = jax_func(jnp.asarray(np_inputs, jnp.bfloat16))
@@ -278,7 +311,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
       def __init__(self):
         super().__init__(name=None)
         if not self.cache:
-          with tf2jax.override_config("force_const_float32_to_bfloat16", True):
+          with config.override_config("force_const_float32_to_bfloat16", True):
             self.cache.append(jax.jit(
                 tf2jax.convert_functional(tf_func, np_inputs)))
 
@@ -344,7 +377,7 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
       tf_outputs = tf_func(tf_inputs)
       tf_grads = tape.gradient(tf_outputs, tf_inputs)
 
-    with tf2jax.override_config("convert_custom_gradient", use_custom_gradient):
+    with config.override_config("convert_custom_gradient", use_custom_gradient):
       jax_func = tf2jax.convert_functional(tf_func, np.zeros_like(np_inputs))
     jax_func = self.variant(jax_func)
     jax_outputs = jax_func(np_inputs)
@@ -422,6 +455,36 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
 
     self.assertAllClose(tf_outputs, jax_outputs)
 
+  def test_metadata(self):
+    @tf.function
+    def tf_func(x, y, *, z):
+      return x + y + z
+
+    x = np.zeros((10, 5), np.float32)
+    y = np.zeros((10, 1), np.float32)
+    z = np.zeros((), np.float32)
+    jax_func = tf2jax.convert_functional(tf_func, x, y=y, z=z)
+    self.assertEqual(jax_func.structured_inputs, (
+        (
+            tf.TensorSpec(shape=(10, 5), dtype=tf.float32, name="x"),
+            tf.TensorSpec(shape=(10, 1), dtype=tf.float32, name="y"),
+        ),
+        {
+            "z": tf.TensorSpec(shape=(), dtype=tf.float32, name="z")
+        },
+    ))
+    self.assertEqual(
+        jax_func.structured_outputs,
+        tf.TensorSpec(shape=(10, 5), dtype=tf.float32, name="Identity"))
+
+    expected_sig = inspect.Signature([
+        inspect.Parameter("x", inspect.Parameter.POSITIONAL_OR_KEYWORD),
+        inspect.Parameter("y", inspect.Parameter.POSITIONAL_OR_KEYWORD),
+        inspect.Parameter("z", inspect.Parameter.KEYWORD_ONLY),
+    ])
+    self.assertEqual(expected_sig, jax_func.signature)
+    self.assertEqual(expected_sig, inspect.signature(jax_func))
+
   @chex.variants(with_jit=True, without_jit=True)
   def test_jit_nesting(self):
     @tf.function
@@ -436,6 +499,24 @@ class FeaturesTest(tf.test.TestCase, parameterized.TestCase):
     inputs = np.linspace(-1., 1., 6, dtype=np.float32).reshape((2, 3))
     self.assertAllClose(tf.sin(tf_fn(inputs)), tf2jax_fn(inputs))
 
+  @chex.variants(with_jit=True, without_jit=True)
+  def test_non_unique_variable_names(self):
+    model = TestMLP(input_size=5, sizes=[7, 8])
+    self.assertNotEqual(
+        len(set([v.name for v in model.variables])), len(model.variables))
+
+    @tf.function
+    def forward(x):
+      return model(x)
+
+    x = np.arange(10).reshape(2, 5).astype(np.float32)
+    tf_outputs = forward(x)
+
+    apply_fn, params = tf2jax.convert(forward, tf.TensorSpec((None, 5)))
+    jax_outputs, _ = self.variant(apply_fn)(params, x)
+
+    self.assertAllClose(tf_outputs, jax_outputs)
+
 
 if __name__ == "__main__":
   tf.test.main()
